#define _BSD_SOURCE
#define _USE_GNU

#include <stdio.h>
#include <string.h>
#include <sys/inotify.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <cxmap.h>
#include <errno.h>
#include <sys/epoll.h>
#include <sys/socket.h>
#include <sys/un.h>

char **parse_interest(const char *pkg)
{
  int paths = 0, i = 0;
  char buf[PATH_MAX];
  FILE *file;
  int rval;
  char **interest;

  rval = snprintf(buf, PATH_MAX, "/var/lib/dpkg/info/%s.triggers", pkg);
  if (rval < 0 || rval == PATH_MAX) {
    return NULL;
  }

  file = fopen(buf, "r");
  if (!file) {
    return NULL;
  }

  while (fgets(buf, PATH_MAX, file)) {
    if (strncmp(buf, "interest ", 9) == 0) {
      ++paths;
    }
  }

  if (paths == 0) {
    fclose(file);
    return NULL;
  }

  rewind(file);
  interest = malloc(sizeof(char *)*(1+paths));
  while (fgets(buf, PATH_MAX, file) && i < paths) {
    if (strncmp(buf, "interest ", 9) == 0) {
      char *src = buf+9, *dst;
      int len = strlen(src);
      if (len > 0 && src[len-1] == '\n') {
	--len;
      }
      dst = malloc(len);
      memcpy(dst, src, len);
      dst[len] = '\0';
      interest[i++] = dst;
    }
  }
  fclose(file);
  interest[i] = NULL;
  return interest;
}

static cx_map *watchpaths;

cxbool less_ptr(cxcptr i1, cxcptr i2)
{
  return i1 < i2;
}

cxbool less_str(cxcptr s1, cxcptr s2)
{
  return strcmp(s1, s2) < 0;
}

int add_watch_recurse(int fd, char *pathname)
{
  DIR *dir = opendir(pathname);
  struct dirent *ent;
  int rval, status;
  size_t pathlen;
  cx_map_iterator cx_iter;

  rval = inotify_add_watch(fd, pathname, IN_CREATE|IN_DELETE|IN_MODIFY|IN_MOVE);
  if (rval > 0) {
    cx_iter = cx_map_insert(watchpaths, (cxcptr)(size_t)rval, pathname);
  }
  if (rval <= 0 || cx_iter == NULL) {
    free(pathname);
    return -1;
  }

  pathlen = strlen(pathname);
  while ((ent = readdir(dir))) {
    if (ent->d_name[0] != '.' && (ent->d_type == DT_UNKNOWN || ent->d_type == DT_DIR)) {
      size_t filelen = strlen(ent->d_name);
      char *newpath = malloc(pathlen+filelen+2);
      rval = 0;
      memcpy(newpath, pathname, pathlen);
      newpath[pathlen] = '/';
      memcpy(newpath+pathlen+1, ent->d_name, filelen+1);
      if (ent->d_type == DT_UNKNOWN) {
	struct stat buf;
	lstat(newpath, &buf);
	if ((buf.st_mode&S_IFDIR)) {
	  rval = add_watch_recurse(fd, newpath);
	} else {
	  free(newpath);
	}
      } else {
	rval = add_watch_recurse(fd, newpath);
      }
      if (!status) {
	status = rval;
      }
    }
  }
  closedir(dir);
  return status;
}

int main()
{
  int optval = 1, inotify_fd, listen_fd, comm_fd = -1, epoll_fd;
  char **interest = parse_interest("man-db"), **ptr;
  struct sockaddr_un sockaddr = {AF_UNIX, "\0inotify-interest"};
  struct epoll_event epoll_event = {EPOLLIN};
  cx_map *changedfiles = cx_map_new(less_str, free, NULL);

  if (!interest) {
    exit(0);
  }
  ptr = interest;
  inotify_fd = inotify_init();
  watchpaths = cx_map_new(less_ptr, NULL, free);
  while (*ptr) {
    add_watch_recurse(inotify_fd, *(ptr++));
  }
  listen_fd = socket(AF_UNIX, SOCK_SEQPACKET, 0);
  bind(listen_fd, (struct sockaddr *)&sockaddr, sizeof(struct sockaddr_un));
  listen(listen_fd, 1);

  epoll_fd = epoll_create(2);
  epoll_event.data.fd = inotify_fd;
  epoll_ctl(epoll_fd, EPOLL_CTL_ADD, inotify_fd, &epoll_event);
  epoll_event.data.fd = listen_fd;
  epoll_ctl(epoll_fd, EPOLL_CTL_ADD, listen_fd, &epoll_event);

  while (epoll_wait(epoll_fd, &epoll_event, 1, -1) == 1 || errno == EINTR) {
    char buf[8192];
    ssize_t nread;
    if (errno == EINTR) {
      errno = 0;
      continue;
    }
    if (epoll_event.data.fd == inotify_fd) {
      int pos = 0;

      nread = read(inotify_fd, buf, 8192);
      while (pos < nread) {
	struct inotify_event *event = (void *)buf+pos;
	if (event->wd == -1) {
	  exit(0);
	}
	if (event->len) {
	  char *idx = rindex(event->name, '.');
	  if (idx && strncmp(idx, ".dpkg-new", 9) == 0) {
	    pos += sizeof(struct inotify_event)+event->len;
	    continue;
	  }
	}

	cx_map_iterator cx_iter = cx_map_find(watchpaths, (cxptr)(size_t)event->wd);
	const char *path = (const char *)cx_map_get_value(watchpaths, cx_iter);

	pos += sizeof(struct inotify_event)+event->len;
	if ((event->mask & IN_IGNORED)) {
	  cx_map_erase_position(watchpaths, cx_iter);
	} else {
	  int len = strlen(path);
	  char *newpath = malloc(len+1+event->len);
	  memcpy(newpath, path, len);
	  newpath[len] = '/';
	  memcpy(newpath+1+len, event->name, event->len);
	  if ((event->mask & IN_ISDIR)) {
	    add_watch_recurse(inotify_fd, newpath);
	  } else {
	    cx_map_insert(changedfiles, newpath, NULL);
	  }
	}
      }
    } else if (epoll_event.data.fd == listen_fd) {
      char anc[1024], buf[4096];
      struct cmsghdr *cmsg;
      struct iovec iov = {.iov_base = buf, .iov_len = sizeof(buf)};
      struct msghdr msg = {.msg_iov = &iov, .msg_iovlen = 1,
			   .msg_control = anc, .msg_controllen = sizeof(anc)};
      comm_fd = accept(listen_fd, NULL, NULL);
      setsockopt(comm_fd, SOL_SOCKET, SO_PASSCRED, &optval, sizeof(optval));
      ssize_t nlen = recvmsg(comm_fd, &msg, 0);
      if (nlen > 0) {
	cx_map_iterator iter, end;
	// Restrict access to root
	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg != NULL; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
	  if (cmsg->cmsg_level == SOL_SOCKET && cmsg->cmsg_type == SCM_CREDENTIALS
	      && ((struct ucred *)CMSG_DATA(cmsg))->uid != 0) {
	    goto next;
	  }
	}
	nlen = sprintf(buf, "%d\n", cx_map_size(changedfiles));
	send(comm_fd, buf, nlen+1, MSG_EOR);
	end = cx_map_end(changedfiles);
	for (iter = cx_map_begin(changedfiles); iter != end;
	     iter = cx_map_next(changedfiles, iter)) {
	  char *val = cx_map_get_key(changedfiles, iter);
	  int len = strlen(val);
	  if (len > 1022) {
	    len = 1022;
	  }
	  memcpy(buf, val, len);
	  buf[len++] = '\n';
	  buf[len++] = '\0';
	  send(comm_fd, buf, len, MSG_EOR);
	}
	cx_map_clear(changedfiles);
      }
    next:
      close(comm_fd);
    }
  }
}
